import logging
from .contexts import ContextualAdapter


def add_logging_level(level_name, level_num):
    """
    Adapted from https://stackoverflow.com/a/35804945
    (Which is to say, shamelessly stolen and tidied up a bit)

    This function will add an additional logging level
    """
    method_name = level_name.lower()
    if hasattr(logging, level_name) and logging.__dict__[level_name] != level_num:
        raise AttributeError(
            "{} already defined in logging module at level {}".format(
                level_name, logging.__dict__[level_name]
            )
        )

    def log_func(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def add_log(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_func)
    setattr(logging, method_name, add_log)

    def log_class_method(self, message, *args, **kwargs):
        self.log(level_num, message, *args, **kwargs)

    setattr(ContextualAdapter, method_name, log_class_method)



from io import StringIO
import json
import logging
import os

import cazoo_logger
from cazoo_logger import add_logging_level
from cazoo_logger import lambda_support as ls
from . import LambdaContext
from pytest import raises

from test.pii_cleaner import PiiFilter


def test_invalid_context_fails():
    event = {"My Fake Event": "data"}
    context = {"My Fake Context": "data"}
    context_type = "bad"

    with raises(Exception):
        ls.LoggerProvider().init_logger(event, context, context_type)


def test_decorated_function_creates_logs():
    event = {"source": "test_event", "detail-type": "test event", "id": "12345"}

    request_id = "abc-123"
    function_name = "testing-the-decorator"
    function_version = "brand-new"

    ctx = LambdaContext(request_id, function_name, function_version)
    stream = StringIO()

    @ls.exception_logger("cloudwatch")
    def handler(event, context, logger):
        cazoo_logger.config(stream)
        logger.info("Logging a test message", extra={"vrm": "LP12 KZM"})

        return "Hello world"

    assert handler(event, ctx) == "Hello world"

    result = json.loads(stream.getvalue())
    assert result["msg"] == "Logging a test message"
    assert result["data"]["vrm"] == "LP12 KZM"
    assert result["level"] == "info"


def test_add_logging_level():
    event = {"source": "test_event", "detail-type": "test event", "id": "12345"}

    request_id = "abc-123"
    function_name = "testing-the-decorator"
    function_version = "brand-new"

    ctx = LambdaContext(request_id, function_name, function_version)
    stream = StringIO()

    os.environ["LOG_LEVEL"] = "TRACE"
    add_logging_level("TRACE", 15)

    @ls.exception_logger("cloudwatch")
    def handler(event, context, logger):
        cazoo_logger.config(stream, level=os.environ["LOG_LEVEL"])
        logger.trace("Logging a test message", extra={"vrm": "LP12 KZM"})
        return "Hello world"

    assert handler(event, ctx) == "Hello world"

    # CLEAN UP THE NEW LOG LEVEL SO IT DOESN'T LEAK INTO OTHER TESTS
    delattr(logging, "TRACE")

    result = json.loads(stream.getvalue())

    assert result["msg"] == "Logging a test message"
    assert result["data"]["vrm"] == "LP12 KZM"
    assert result["level"] == "trace"


def test_handler_logger_with_filter():
    event = {"source": "test_event", "detail-type": "test event", "id": "12345"}

    request_id = "abc-123"
    function_name = "testing-the-decorator"
    function_version = "brand-new"

    ctx = LambdaContext(request_id, function_name, function_version)
    stream = StringIO()

    @ls.handler_logger("cloudwatch", log_filter=PiiFilter)
    def handler(event, context, logger):
        cazoo_logger.config(stream)
        logger.info(
            "Logging a test message",
            extra={"vrm": "LP12 KZM", "email_address": "me@email.com"},
        )

        return "Hello world"

    assert handler(event, ctx) == "Hello world"

    result = json.loads(stream.getvalue())
    assert result["msg"] == "Logging a test message"
    assert result["data"]["vrm"] == "LP12 KZM"
    assert result["level"] == "info"
    assert result["data"] == {"vrm": "LP12 KZM", "email_address": "PII REMOVED"}


def test_handler_logger_removes_pii_empty():
    event = {"source": "test_event", "detail-type": "test event", "id": "12345"}

    request_id = "abc-123"
    function_name = "testing-the-decorator"
    function_version = "brand-new"

    ctx = LambdaContext(request_id, function_name, function_version)
    stream = StringIO()

    @ls.handler_logger("empty", log_filter=PiiFilter)
    def handler(event, context, logger):
        cazoo_logger.config(stream)
        logger.info(
            "Logging a test message",
            extra={"vrm": "LP12 KZM", "email_address": "me@email.com"},
        )

        return "Hello world"

    assert handler(event, ctx) == "Hello world"

    result = json.loads(stream.getvalue())
    assert result["data"] == {"vrm": "LP12 KZM", "email_address": "PII REMOVED"}



import logging
from ._version import get_versions
from .formatters import JsonFormatter
from . import contexts
from .logging_levels import add_logging_level
from collections import ChainMap

__version__ = get_versions()["version"]
del get_versions

__all__ = ["empty", "s3", "cloudwatch", "config", "add_logging_level"]


def s3(event, context, service=None):
    """
    Build a contextual logger for an S3 SNS notification.
    """
    return contexts.S3SnsContext(event, context, logging.root, service)


def cloudwatch(event, context, service=None):
    """
    Build a contextual logger for a Cloudwatch event
    """
    return contexts.CloudwatchContext(event, context, logging.root, service)


def config(stream=None, level=logging.INFO, boto_level=logging.WARN):
    stdout = logging.StreamHandler(stream)
    stdout.setLevel(level)
    stdout.setFormatter(JsonFormatter())
    logging.root.setLevel(level)
    logging.root.handlers.clear()
    logging.root.addHandler(stdout)
    logging.getLogger("boto").setLevel(boto_level)
    logging.getLogger("botocore").setLevel(boto_level)
    logging.getLogger("boto3").setLevel(boto_level)


def empty():
    """
    Return an empty logger, for use in unit tests or non-lambda envs.
    """
    return contexts.ContextualAdapter(logging.root, ChainMap())



import json
import logging


def json_formatter(obj):
    """request_id"""
    return str(obj)


class JsonFormatter(logging.Formatter):
    """AWS Lambda Logging formatter."""

    def __init__(self, **kwargs):
        """Return a JsonFormatter instance.

        The `json_default` kwarg is used to specify a formatter for otherwise
        unserialisable values.  It must not throw.  Defaults to a function that
        coerces the value to a string.
        """
        datefmt = kwargs.pop("datefmt", None)

        super(JsonFormatter, self).__init__(datefmt=datefmt)
        self.default_json_formatter = kwargs.pop("json_default", json_formatter)

        self._supported = {"level", "context", "data", "type"}

    def format(self, record):
        record_dict = record.__dict__.copy()
        record_dict["asctime"] = self.formatTime(record, self.datefmt)

        log_dict = {k: v for k, v in record_dict.items() if k in self._supported and v}
        log_dict["msg"] = record.getMessage()
        log_dict["level"] = record.levelname.lower()

        if record.exc_info:
            exc_type, exc, exc_info = record.exc_info
            err = {
                "name": exc_type.__name__,
                "message": str(exc),
                "stack": self.formatException(record.exc_info),
            }

            if "data" not in log_dict:
                log_dict["data"] = {"error": err}
            else:
                log_dict["data"]["error"] = err

        json_record = json.dumps(log_dict, default=self.default_json_formatter)

        if hasattr(json_record, "decode"):  # pragma: no cover
            json_record = json_record.decode("utf-8")

        return json_record



"""
LAMBDA SUPPORT
This module provides useful functions for implementing the logger in Lambda functions.

LoggerProvider
A wrapper class that allows the logger to be easily instantiated and pulled into code
init_logger - will create a logger instance if one does not already exist
get_logger - will retrieve an existing logger instance

exception_logger
Is used as a decorator on the root Lambda Handler in a Lambda function. The function
will create an instance of the logger and then log the incoming event data.

If there are any unhandled exceptions when the handler is run then the Exception detail
is logged.  The Exception is still raised to ensure that the Lambda is correctly marked
as a failure

handler_logger
Is used as a decorator on the root Lambda Handler in a Lambda function. The function
will create an instance of the logger and then log the incoming event data.

Accepts a prelog_hook function that will be applied to all additional data logged
e.g. this can be used to remove any PII that is accidentally logged
"""

import functools
import os

from . import config, cloudwatch, empty
from .contexts import CloudwatchContext, ContextualAdapter


class LoggerProvider:
    """
    Helper class for instantiating the Cazoo Logger.  This will be called when the new
    handler decorator is fired, and so we want to ensure that the logger is reset each
    time.
    """

    logger = None

    @staticmethod
    def init_logger(
        event, context, context_type
    ) -> [CloudwatchContext, ContextualAdapter]:

        config(level=os.environ.get("LOG_LEVEL", "INFO"))
        if context_type == "cloudwatch":
            LoggerProvider.logger = cloudwatch(event, context)
        elif context_type == "empty":
            LoggerProvider.logger = empty()
        else:
            raise Exception("Invalid context type {0}".format(context_type))
        return LoggerProvider.logger

    @staticmethod
    def get_logger() -> [CloudwatchContext, ContextualAdapter]:
        return LoggerProvider.logger


def exception_logger(context_type, has_pii=False):
    """
    Decorator for the Lambda handler that will instantiate the logger and log initial
    event state data.
    :param context_type: The context of the incoming cloudwatch event.
    :param has_pii: Flag to disable blanket event login for cases where events hold
                    pii data that cannot be logged to cloudwatch
    """

    def log_decorator(handler):
        @functools.wraps(handler)
        def exception_handler(event, context):
            log = LoggerProvider.init_logger(event, context, context_type=context_type)
            if not has_pii:
                log.info("Logging event data", extra={"event": event})
            try:
                return handler(event, context, log)
            except Exception:
                extra = {}
                if not has_pii:
                    extra["event"] = event
                log.exception("Unhandled exception in Lambda", extra=extra)
                raise

        return exception_handler

    return log_decorator


def handler_logger(context_type, log_filter=None):
    """
    Decorator for the Lambda handler that will instantiate the logger and log initial
    event state data.
    :param context_type: The context of the incoming cloudwatch event.
    :param log_filter: An optional LogFilter that will allow log data to be pre-cleaned
                       e.g. to remove PII
    """

    def log_decorator(handler):
        @functools.wraps(handler)
        def exception_handler(event, context):
            log = LoggerProvider.init_logger(event, context, context_type=context_type)
            if log_filter:
                log.addFilter(log_filter())
            log.info("Logging event data", extra={"event": event})
            try:
                return handler(event, context, log)
            except Exception:
                log.exception("Unhandled exception in Lambda", extra=event)
                raise

        return exception_handler

    return log_decorator



import logging
from collections import ChainMap


class ContextualAdapter(logging.LoggerAdapter):
    def __init__(self, logger, data=None):
        self.context = data
        super().__init__(logger, data)

    def with_context(self, **ctx):
        new_ctx = self.context.new_child()
        new_ctx.update({"context": ctx})
        return ContextualAdapter(self.logger, new_ctx)

    def with_data(self, **ctx):
        new_ctx = self.context.new_child()
        new_ctx.update({"data": ctx})
        return ContextualAdapter(self.logger, new_ctx)

    def process(self, msg, kwargs):
        if "extra" in kwargs:
            extra = kwargs["extra"].copy()
            del kwargs["extra"]
            kwargs["extra"] = {"data": extra}
            kwargs["extra"].update(self.context)
        else:
            kwargs["extra"] = self.context

        if "type" in kwargs:
            kwargs["extra"]["type"] = kwargs.pop("type")

        return msg, kwargs

    def addFilter(self, filter):
        self.logger.addFilter(filter)


class LambdaContext(ContextualAdapter):
    def __init__(self, context, data, logger, service=None):
        default = {
            "context": {
                "request_id": context.aws_request_id,
                "function": {
                    "name": context.function_name,
                    "version": context.function_version,
                },
            }
        }
        if service is not None:
            default["context"]["function"]["service"] = service
        default["context"].update(data)
        if service is not None:
            default["context"]["function"]["service"] = service
        super().__init__(logger, ChainMap(default))


class S3SnsContext(LambdaContext):
    def __init__(self, event, context, logger, service=None):
        [record] = event["Records"]
        super().__init__(
            context,
            {
                "sns": {
                    "id": record["Sns"]["MessageId"],
                    "type": record["Sns"]["Type"],
                    "topic": record["Sns"]["TopicArn"],
                    "subject": record["Sns"]["Subject"],
                }
            },
            logger,
            service,
        )


class CloudwatchContext(LambdaContext):
    def __init__(self, event, context, logger, service=None):
        super().__init__(
            context,
            {
                "event": {
                    "source": event["source"],
                    "name": event["detail-type"],
                    "id": event["id"],
                }
            },
            logger,
            service,
        )



from setuptools import setup
import versioneer
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="cazoo_logger",
    version=versioneer.get_version(),
    url="https://github.com/Cazoo-uk/py-logger",
    license="MIT",
    author="Bob Gregory",
    author_email="bob.gregory@cazoo.co.uk",
    description="Super-opinionated structured logger for AWS lambda",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    packages=["cazoo_logger"],
    cmdclass=versioneer.get_cmdclass(),
)
